QuestsFu = FuBarPlugin:new({
	name          = QuestsFuLocals.NAME,
	description   = QuestsFuLocals.DESCRIPTION,
	version       = "0.2.1.1",
	releaseDate   = "04-08-2006",
	aceCompatible = 103,
	fuCompatible  = 910,
	author        = "Kemayo",
	email         = "kemayo@gmail.com",
	website       = "http://kemayo.wowinterface.com/",
	category      = "quests",
	db            = AceDatabase:new("QuestsFuDB"),
	defaults      = {
		showLevels = TRUE,
		showDifficulty = TRUE,
		showArea = TRUE,
		showCurrentAreaOnly = FALSE,
		showCurrentAreaDescriptionOnly = FALSE,
		showDescription = TRUE,
		showCompletedObjectives = FALSE,
		showInItemTooltip = TRUE,
		colorObjectives = TRUE,
		showText = {
			total = TRUE,
			complete = FALSE,
		},
	},
	charDefaults  = {
		hidden = {}
	},
	cmd           = AceChatCmd:new(QuestsFuLocals.COMMANDS, QuestsFuLocals.CMD_OPTIONS),
	loc           = QuestsFuLocals,
	hasIcon       = TRUE,
	clickableTooltip = TRUE,
	
	-- Methods
	
	ToggleShowingText = function(self, setting, loud)
		self.data.showText[setting] = not self.data.showText[setting]
		self:UpdateText()
		return self.data.showText[setting]
	end,
	
	IsShowingText = function(self, setting)
		return self.data.showText[setting]
	end,
	
	IsShowingLevels = function(self)
		return self.data.showLevels
	end,
	
	ToggleShowingLevels = function(self, silent)
		self.data.showLevels = not self.data.showLevels
		self:UpdateTooltip()
		return self.data.showLevels
	end,
	
	IsShowingDifficulty = function(self)
		return self.data.showDifficulty
	end,
	
	ToggleShowingDifficulty = function(self, silent)
		self.data.showDifficulty = not self.data.showDifficulty
		self:UpdateTooltip()
		return self.data.showDifficulty
	end,
	
	IsShowingArea = function(self)
		return self.data.showArea
	end,
	
	ToggleShowingArea = function(self, silent)
		self.data.showArea = not self.data.showArea
		self:UpdateTooltip()
		return self.data.showArea
	end,
	
	IsShowingCurrentAreaOnly = function(self)
		return self.data.showCurrentAreaOnly
	end,
	
	ToggleShowingCurrentAreaOnly = function(self, silent)
		self.data.showCurrentAreaOnly = not self.data.showCurrentAreaOnly
		self:UpdateTooltip()
		return self.data.showCurrentAreaOnly
	end,
	
	IsShowingCurrentAreaDescriptionOnly = function(self)
		return self.data.showCurrentAreaDescriptionOnly
	end,
	
	ToggleShowingCurrentAreaDescriptionOnly = function(self, silent)
		self.data.showCurrentAreaDescriptionOnly = not self.data.showCurrentAreaDescriptionOnly
		self:UpdateTooltip()
		return self.data.showCurrentAreaDescriptionOnly
	end,
	
	IsShowingDescription = function(self)
		return self.data.showDescription
	end,
	
	ToggleShowingDescription = function(self, silent)
		self.data.showDescription = not self.data.showDescription
		self:UpdateTooltip()
		return self.data.showDescription
	end,
	
	IsColoringObjectives = function(self)
		return self.data.colorObjectives
	end,
	
	ToggleColoringObjectives = function(self, silent)
		self.data.colorObjectives = not self.data.colorObjectives
		self:UpdateTooltip()
		return self.data.colorObjectives
	end,
	
	IsShowingCompletedObjectives = function(self)
		return self.data.showCompletedObjectives
	end,
	
	ToggleShowingCompletedObjectives = function(self, silent)
		self.data.showCompletedObjectives = not self.data.showCompletedObjectives
		self:UpdateTooltip()
		return self.data.showCompletedObjectives
	end,
	
	IsShowingInItemTooltip = function(self)
		return self.data.showInItemTooltip
	end,
	
	ToggleShowingInItemTooltip = function(self, silent)
		self.data.showInItemTooltip = not self.data.showInItemTooltip
		self:UpdateTooltip()
		return self.data.showDescription
	end,
	
	Initialize = function(self)
		--In case of upgrade, handle anything where the type has changed.
		--if self.data.version ~= self.versionNumber then
		--
		--end
		
		self.numQuests = 0
		self.numEntries = 0
		self.quests = {}
		self.items = {}
		self.mobs = {}
	end,
	
	Enable = function(self)
		DEFAULT_CHAT_FRAME:AddMessage("FuQ enabled.")
		self:RegisterEvent("QUEST_LOG_UPDATE", "Update")
		self:RegisterEvent("PLAYER_LEVEL_UP", "Update") -- Quest difficulty colors can change on level
		self:RegisterEvent("UPDATE_FACTION", "Update") -- For reputation quests (is this entirely neccessary?  A faction update might fire QUEST_LOG_UPDATE...)
		self:RegisterEvent("UPDATE_MOUSEOVER_UNIT", "OnMouseOverUnit")
		self:Hook("ContainerFrameItemButton_OnEnter", "OnItemTooltip")
	end,
	
	OnMouseOverUnit = function(self)
		-- Make sure that this is wanted, and that the tooltip is actually read/writeable.
		if self:IsShowingInItemTooltip() and self:CanMessWithGameTooltip() then
			--Comes with color codes.  Strip 'em out.
			local thisMob = string.gsub(getglobal('GameTooltipTextLeft1'):GetText(),"|c........(.*)|?r?","%1")
			if self.mobs[thisMob] then
				self:AddToGameTooltipAndFixHeight(self.mobs[thisMob].quest..": "..self.mobs[thisMob].needed, self.mobs[thisMob].color)
			end
		end
	end,
	
	OnItemTooltip = function(self)
		self.Hooks["ContainerFrameItemButton_OnEnter"].orig()
		if self:IsShowingInItemTooltip() and self:CanMessWithGameTooltip() then
			--Comes with color codes.  Strip 'em out.
			local thisItem = string.gsub(getglobal('GameTooltipTextLeft1'):GetText(),"|c........(.*)|?r?","%1")
			if self.items[thisItem] then
				self:AddToGameTooltipAndFixHeight(self.items[thisItem].quest..": "..self.items[thisItem].needed, self.items[thisItem].color)
			end
		end
	end,
	
	CanMessWithGameTooltip = function(self)
		if GameTooltip and getglobal('GameTooltipTextLeft1'):IsVisible() and getglobal('GameTooltipTextLeft1'):GetText() ~= nil then
			return TRUE
		else
			return FALSE
		end
	end,
	
	AddToGameTooltipAndFixHeight = function(self,text,color)
		GameTooltip:AddLine(text, color.r, color.g, color.b)
		
		-- Fix up the width of the tooltip
		length = getglobal(GameTooltip:GetName() .. "TextLeft" .. GameTooltip:NumLines()):GetStringWidth()
		-- Space for right-border:
		length = length + 22
		
		GameTooltip:SetHeight(GameTooltip:GetHeight() + 14)
		if length > GameTooltip:GetWidth() then
			GameTooltip:SetWidth(length)
		end
	end,
	
	MenuSettings = function(self, level, value)
		if level == 1 then
			UIDropDownMenu_AddButton({
				text = self.loc.MENU_SHOW_LEVELS,
				func = function()
					self:ToggleShowingLevels()
				end,
				checked = self:IsShowingLevels(),
				keepShownOnClick = TRUE,
			})
			
			UIDropDownMenu_AddButton({
				text = self.loc.MENU_SHOW_DIFFICULTY,
				func = function()
					self:ToggleShowingDifficulty()
				end,
				checked = self:IsShowingDifficulty(),
				keepShownOnClick = TRUE,
			})
			
			UIDropDownMenu_AddButton({
				text = self.loc.MENU_SHOW_AREA,
				func = function()
					self:ToggleShowingArea()
				end,
				checked = self:IsShowingArea(),
				keepShownOnClick = TRUE,
			})
			
			UIDropDownMenu_AddButton({
				text = self.loc.MENU_SHOW_CURRENT_AREA_ONLY,
				func = function()
					self:ToggleShowingCurrentAreaOnly()
				end,
				checked = self:IsShowingCurrentAreaOnly(),
				keepShownOnClick = TRUE,
			})
			
			UIDropDownMenu_AddButton({
				text = self.loc.MENU_SHOW_DESCRIPTION,
				func = function()
					self:ToggleShowingDescription()
				end,
				checked = self:IsShowingDescription(),
				keepShownOnClick = TRUE,
			})
			
			UIDropDownMenu_AddButton({
				text = self.loc.MENU_COLOR_OBJECTIVES,
				func = function()
					self:ToggleColoringObjectives()
				end,
				checked = self:IsColoringObjectives(),
				keepShownOnClick = TRUE,
			})
			
			UIDropDownMenu_AddButton({
				text = self.loc.MENU_SHOW_COMPLETED_OBJECTIVES,
				func = function()
					self:ToggleShowingCompletedObjectives()
				end,
				checked = self:IsShowingCompletedObjectives(),
				keepShownOnClick = TRUE,
			})
			
			UIDropDownMenu_AddButton({
				text = self.loc.MENU_SHOW_CURRENT_AREA_DESCRIPTION_ONLY,
				func = function()
					self:ToggleShowingCurrentAreaDescriptionOnly()
				end,
				checked = self:IsShowingCurrentAreaDescriptionOnly(),
				keepShownOnClick = TRUE,
			})
			
			UIDropDownMenu_AddButton({
				text = self.loc.MENU_SHOW_IN_ITEM_TOOLTIP,
				func = function()
					self:ToggleShowingInItemTooltip()
				end,
				checked = self:IsShowingInItemTooltip(),
				keepShownOnClick = TRUE,
			})
			
			UIDropDownMenu_AddButton({
				text = self.loc.MENU_SHOW_TEXT,
				func = function()
					if self:IsShowingText("total") then
						self:ToggleShowingText("total")
					else
						self:ToggleShowingText("complete")
					end
				end,
				checked = self:IsShowingText("total") or self:IsShowingText("complete"),
				hasArrow = self:IsShowingText("total") or self:IsShowingText("complete"),
				value = "total",
			})
		elseif level == 2 then
			if value=="total" then
				UIDropDownMenu_AddButton({
					text = self.loc.MENU_SHOW_TEXT_TOTAL,
					func = function()
						self:ToggleShowingText("total")
						if not self:IsShowingText("total") and not self:IsShowingText("complete") then
							HideDropDownMenu(1)
						end
					end,
					checked = self:IsShowingText("total"),
					keepShownOnClick = TRUE,
				}, level)
				
				UIDropDownMenu_AddButton({
					text = self.loc.MENU_SHOW_TEXT_COMPLETE,
					func = function()
						self:ToggleShowingText("complete")
						if not self:IsShowingText("complete") and not self:IsShowingText("total") then
							HideDropDownMenu(1)
						end
					end,
					checked = self:IsShowingText("complete"),
					keepShownOnClick = TRUE,
				}, level)
			end
		end
	end,
	
	Report = function(self)
		self.cmd:report({
			{text=self.loc.ARGUMENT_SHOWLEVELS, val=(self:IsShowingLevels() and 1 or 0), map=FuBarLocals.MAP_ONOFF},
			{text=self.loc.ARGUMENT_SHOWDIFFICULTY, val=(self:IsShowingDifficulty() and 1 or 0), map=FuBarLocals.MAP_ONOFF},
			{text=self.loc.ARGUMENT_SHOWAREA, val=(self:IsShowingArea() and 1 or 0), map=FuBarLocals.MAP_ONOFF},
			{text=self.loc.ARGUMENT_SHOWCURRENTAREAONLY, val=(self:IsShowingCurrentAreaOnly() and 1 or 0), map=FuBarLocals.MAP_ONOFF},
			{text=self.loc.ARGUMENT_SHOWDESCRIPTION, val=(self:IsShowingDescription() and 1 or 0), map=FuBarLocals.MAP_ONOFF},
			{text=self.loc.ARGUMENT_COLOROBJECTIVES, val=(self:IsColoringObjectives() and 1 or 0), map=FuBarLocals.MAP_ONOFF},
			{text=self.loc.ARGUMENT_COMPLETEDOBJECTIVES, val=(self:IsShowingCompletedObjectives() and 1 or 0), map=FuBarLocals.MAP_ONOFF},
			{text=self.loc.ARGUMENT_CURRENTAREADESCRIPTIONONLY, val=(self:IsShowingCurrentAreaDescriptionOnly() and 1 or 0), map=FuBarLocals.MAP_ONOFF},
			{text=self.loc.ARGUMENT_SHOWINITEMTOOLTIP, val=(self:IsShowingInItemTooltip() and 1 or 0), map=FuBarLocals.MAP_ONOFF},
		})
	end,
	
	UpdateData = function(self)
		--UpdateText and UpdateTooltip both need to know things about the questlog.  Better to do this once.
		local startingQuestLogSelection = GetQuestLogSelection()
		local numEntries, numQuests = GetNumQuestLogEntries()
		local numQuestsDone = 0
		local quests = {}
		local questzones = {}
		local items = {}
		local mobs = {}
		local zoneIndex = "OMG NONE" --The first item in the quest log should always be a header, so this should always be replaced.  But just in case.
		local quest
		
		if numEntries > 0 then
			for quest = 1, numEntries, 1 do
				SelectQuestLogEntry(quest)
				local strQuestLogTitleText, strQuestLevel, strQuestTag, isHeader, isCollapsed, isComplete = GetQuestLogTitle(quest)
				local strQuestDescription, strQuestObjectives = GetQuestLogQuestText()
				local q = {}
				local thisQuest = "  "
				local strQuestLevelTag = ""
				
				if strQuestTag == self.loc.ELITE then
					strQuestLevelTag = "+"
				elseif strQuestTag == self.loc.DUNGEON then
					strQuestLevelTag = "D"
				elseif strQuestTag == self.loc.PVP then
					strQuestLevelTag = "P"
				elseif strQuestTag == self.loc.RAID then
					strQuestTag = "R"
				end
				
				if isHeader then
					zoneIndex = strQuestLogTitleText
					table.insert(questzones, zoneIndex)
					quests[zoneIndex] = {}
				else
					if isComplete == 1 then
						isComplete = format("(%s)",self.loc.QUEST_DONE)
						numQuestsDone = numQuestsDone + 1
					elseif isComplete == -1 then
						isComplete = format("(%s)",self.loc.QUEST_FAILED)
					else
						isComplete = nil
					end
					
					q.title = strQuestLogTitleText
					q.level = strQuestLevel
					q.tag = strQuestLevelTag
					q.complete = isComplete
					q.objective = strQuestObjectives
					
					local lb = {}
					if (GetNumQuestLeaderBoards() > 0) then
						for ii=1, GetNumQuestLeaderBoards() do
							local desc, qtype, done = GetQuestLogLeaderBoard(ii)
							
							local itemName,mobName,numNeeded,numItems,factionName,factionCurrent,factionNeeded
							if qtype == 'item' then
								itemName,numItems,numNeeded = BabbleLib:GetInstance("1.0").Deformat(desc, QUEST_OBJECTS_FOUND)
								desc = itemName
							elseif qtype == 'monster' then
								mobName,numItems,numNeeded = BabbleLib:GetInstance("1.0").Deformat(desc, QUEST_MONSTERS_KILLED)
								if mobName == nil or numItems == nil or numNeeded == nil then
									--Sometimes we get objectives like "Find Mankrik's Wife: 0/1", which are listed as "monster".
									mobName, numItems, numNeeded = BabbleLib:GetInstance("1.0").Deformat(desc, QUEST_OBJECTS_FOUND)
								end
								desc = mobName
							elseif qtype == 'reputation' then
								factionName,factionCurrent,factionNeeded = BabbleLib:GetInstance("1.0").Deformat(desc, QUEST_FACTION_NEEDED)
								numItems = self:GetReactionLevel(factionCurrent)
								numNeeded = self:GetReactionLevel(factionNeeded)
								desc = factionName
							end
							
							local color = {}
							if numItems ~= nil then
								-- This quest involves items!
								color.r, color.g, color.b = FuBarUtils.GetThresholdColor(numItems / numNeeded)
							else
								-- This quest doesn't involve items.  Let's leave it grey.  We're lazy.
								color = {r=0.5,g=0.5,b=0.5}
							end
							
							if qtype == 'item' then
								items[itemName] = {quest = strQuestLogTitleText, item = itemName, needed = string.format("%s/%s",numItems,numNeeded), color=color}
							elseif qtype == 'monster' then
								mobs[mobName] = {quest = strQuestLogTitleText, mob = mobName, needed = string.format("%s/%s",numItems,numNeeded), color=color}
							end
							
							if done then
								done = format("(%s)",self.loc.QUEST_DONE)
							elseif numItems and numNeeded then
								if qtype == 'reputation' then
									done = string.format("%s/%s",factionCurrent,factionNeeded)
								else
									done = string.format("%s/%s",numItems,numNeeded)
								end
							else
								--I think this shouldn't happen.  ...but I haven't seen every quest out there.
								done = ""
							end
							
							lb[ii] = {description=desc, done=done, color=color}
						end
					end
					q.leaderboard = lb
					q.id = quest
					
					table.insert(quests[zoneIndex],q)
				end
			end
		end
		
		--table.sort(quests, function(a,b) return string.byte(a) < string.byte(b) end)
		self.quests = quests
		self.questzones = questzones
		self.items = items
		self.mobs = mobs
		self.numQuests = numQuests
		self.numEntries = numEntries
		self.numQuestsDone = numQuestsDone
		
		SelectQuestLogEntry(startingQuestLogSelection)
	end,
	
	UpdateText = function(self)
		local text = ""
		if self:IsShowingText("total") then
			local maxQuests = 20
			local color = FuBarUtils.GetThresholdHexColor((maxQuests - self.numQuests) / maxQuests)
			text = text .. FuBarUtils.Colorize(color,format("%d/%d", self.numQuests, maxQuests))
		end
		if self:IsShowingText("total") and self:IsShowingText("complete") then
			text = text .. " c:"
		end
		if self:IsShowingText("complete") then
			local maxQuests = self.numQuests
			local color = FuBarUtils.GetThresholdHexColor(self.numQuestsDone / maxQuests)
			text = text .. FuBarUtils.Colorize(color,format("%d/%d", self.numQuestsDone, maxQuests))
		end
		self:SetText(text)
	end,
	
	UpdateTooltip = function(self)
		local currentZone = GetZoneText()
		local cat = nil
		self.numBut = 0
		
		if self.numQuests > 0 then
			if table.getn(self.questzones) then
				local zone
				for _,zone in self.questzones do
					-- If showing area headers and either not showing only the current area's quests or this is the current area.
					if self:IsShowingArea() and ((currentZone == zone) or not self:IsShowingCurrentAreaOnly()) then
						cat = zone
						-- NOTE TO SELF: On next BossPanel release after 0.9.9.229, change hideBlankLine (the last one) to true.
						FuBarTooltip:AddCategory(cat, FuBarUtils.Colorize("bfffbf",zone), nil, "ToggleCategory", nil, false)
					end
					-- If we're not hiding the zone and it's either the current zone or we're not showing only the current zone.
					if not self.charData.hidden[zone] and ((currentZone == zone) or not self:IsShowingCurrentAreaOnly()) then
						if table.getn(self.quests[zone]) > 0 then
							local quest
							for _,quest in self.quests[zone] do
								local thisQuest
								local questLevel = ""
								
								if self:IsShowingLevels() then
									questLevel = format("[%s%s] ", quest.level, quest.tag)
								end
								
								thisQuest = questLevel..quest.title
								
								local thisQuestColor = {r=1,g=1,b=1}
								if self:IsShowingDifficulty() then
									thisQuestColor = GetDifficultyColor(quest.level)
								end
								
								FuBarTooltip:AddDoubleLine(cat, thisQuest, quest.complete, thisQuestColor.r, thisQuestColor.g, thisQuestColor.b, nil, nil, nil, nil, string.format("%d obj 0", quest.id), "QuestClick", (IsQuestWatched(quest.id) and true or false))
								
								-- If we're showing descriptions/objectives and it's either the current zone or we're not showing descriptions only for the current zone.
								if self:IsShowingDescription() and ((currentZone == zone) or not self:IsShowingCurrentAreaDescriptionOnly()) then
									-- If we know of objectives:
									if table.getn(quest.leaderboard) > 0 then
										local i, goal
										for i,goal in quest.leaderboard do
											if not (string.find(goal.done, self.loc.QUEST_DONE) and not self:IsShowingCompletedObjectives()) then
												local color
												if self:IsColoringObjectives() then
													color = goal.color
												else
													color = thisQuestColor
												end
												FuBarTooltip:AddDoubleLine(cat, '  '..goal.description, goal.done, color.r, color.g, color.b, color.r, color.g, color.b, FuBarTooltip:GetNormalFontSize()-2, string.format("%d obj %d", quest.id, i), "QuestClick")
											end
										end
									-- Otherwise, if the quest is incomplete (or complete and we're showing completed objectives.
									elseif not (quest.complete and not self:IsShowingCompletedObjectives()) then
										FuBarTooltip:AddLine(cat, '  '..quest.objective, 0.5, 0.5, 0.5, TRUE, "LEFT", FuBarTooltip:GetNormalFontSize()-2, string.format("%d obj 0", quest.id), "QuestClick")
									end
								end
							end
						end
					end
				end
			end
		end
		FuBarTooltip:SetHint("Click to open the quest log")
	end,
	
	ToggleCategory = function(self, id, button)
		if self.charData.hidden[id] then
			self.charData.hidden[id] = FALSE
		else
			self.charData.hidden[id] = TRUE
		end
		-- Refresh in place
		self:UpdateTooltip()
	end,
	
	--TheFly contributed the base of this, and it's still using much of the logic he provided
	QuestClick = function(self, id, button)
		local _,_,questid = string.find(id, "(.+) obj .+")
		questid = tonumber(questid)
		if IsShiftKeyDown() then
			-- Shift-click toggles quest-watch on this quest.
			if IsQuestWatched(questid) then
				RemoveQuestWatch(questid)
				QuestWatch_Update()
			else
				-- Set error if no objectives
				if GetNumQuestLeaderBoards(questid) == 0 then
					UIErrorsFrame:AddMessage(QUEST_WATCH_NO_OBJECTIVES, 1.0, 0.1, 0.1, 1.0, UIERRORS_HOLD_TIME)
					return
				end
				-- Set an error message if trying to show too many quests
				if ( GetNumQuestWatches() >= MAX_WATCHABLE_QUESTS ) then
					UIErrorsFrame:AddMessage(format(QUEST_WATCH_TOO_MANY, MAX_WATCHABLE_QUESTS), 1.0, 0.1, 0.1, 1.0, UIERRORS_HOLD_TIME)
					return
				end
				AddQuestWatch(questid)
				QuestWatch_Update()
			end
			FuBarTooltip:ToggleChecked(string.format("%d obj 0", questid))
		else
			if QuestLogFrame:IsVisible() then
				if self.lastIndex == questid then
					HideUIPanel(QuestLogFrame)
				end
			else
				ShowUIPanel(QuestLogFrame)
			end
			if (self.numEntries > QUESTS_DISPLAYED) then
				if (questid < self.numEntries - QUESTS_DISPLAYED) then
					FauxScrollFrame_SetOffset(QuestLogListScrollFrame, questid - 1)
					QuestLogListScrollFrameScrollBar:SetValue((questid - 1) * QUESTLOG_QUEST_HEIGHT)
				else
					FauxScrollFrame_SetOffset(QuestLogListScrollFrame, self.numEntries - QUESTS_DISPLAYED)
					QuestLogListScrollFrameScrollBar:SetValue((self.numEntries - QUESTS_DISPLAYED) * QUESTLOG_QUEST_HEIGHT)
				end
			end
			
			SelectQuestLogEntry(questid)
			QuestLog_SetSelection(questid)
			self.lastIndex = questid
		end
	end,
	
	OnClick = function(self)
		ToggleQuestLog()
	end,
	
	GetReactionLevel = function(self, leveltext)
		if leveltext == FACTION_STANDING_LABEL1 then --Hated
			return 1
		elseif leveltext == FACTION_STANDING_LABEL2 then --Hostile
			return 2
		elseif leveltext == FACTION_STANDING_LABEL3 then --Unfriendly
			return 3
		elseif leveltext == FACTION_STANDING_LABEL4 then --Neutral
			return 4
		elseif leveltext == FACTION_STANDING_LABEL5 then --Friendly
			return 5
		elseif leveltext == FACTION_STANDING_LABEL6 then --Honored
			return 6
		elseif leveltext == FACTION_STANDING_LABEL7 then --Revered
			return 7
		elseif leveltext == FACTION_STANDING_LABEL8 then --Exalted
			return 8
		end
	end,
})

QuestsFu:RegisterForLoad()
